import argparse
import os
import re

from linkmedic.__init__ import __version__


class Orders:
    """Diagnostic orders for the link medic"""

    internal_check = False
    external_check = False
    http_check = False
    write_badge = False
    ignore_status = set()
    ignore_urls = set()

    def __init__(self, args_namespace):
        self.internal_check = not args_namespace.ignore_local
        self.external_check = args_namespace.check_external
        self.http_check = args_namespace.warn_http
        self.write_badge = args_namespace.with_badge
        self.ignore_status = set(args_namespace.ignore_status)

        if not (self.internal_check or self.external_check):
            self.external_check = True

        LINKS_IGNORE_FILE_PATH = os.path.join(args_namespace.root, ".linkignore")
        LINKS_IGNORE_FILE_FOUND = os.path.isfile(LINKS_IGNORE_FILE_PATH)
        if LINKS_IGNORE_FILE_FOUND:
            with open(
                LINKS_IGNORE_FILE_PATH, "r", encoding="utf-8"
            ) as linksignore_file:
                for ignore_item in linksignore_file.read().splitlines():
                    sanitized_ignore_item = ignore_item.strip().lower()
                    if sanitized_ignore_item[0] != "/" and not re.match(
                        r"^https?:\/{2}.*", sanitized_ignore_item
                    ):
                        sanitized_ignore_item = "/" + sanitized_ignore_item
                    self.ignore_urls.add(sanitized_ignore_item)


def cli_args():
    """returns cli arguments"""
    parser = argparse.ArgumentParser(description="Simple python website links checker")
    parser.add_argument(
        "-r",
        "--root",
        default="./",
        help="path of the webserver's root folder (default=./)",
    )
    parser.add_argument(
        "-e",
        "--entry",
        default="index.html",
        help="path of the entry page on the webserver, relative to the root (default=index.html)",
    )
    parser.add_argument("--verbose", "-v", action="count", default=0)
    parser.add_argument("--quiet", action="store_true", help="show only error logs")
    parser.add_argument("--silent", action="store_true", help="slience the output logs")
    parser.add_argument("--exit-zero", action="store_true", help="always return zero")
    parser.add_argument(
        "--no-redirect",
        action="store_true",
        help="do not redirect the 'missing_page' to 'missing_page.html'",
    )
    parser.add_argument(
        "--check-external", action="store_true", help="check links to external domains"
    )
    parser.add_argument(
        "--ignore-local", action="store_true", help="ignore local dead links"
    )
    parser.add_argument(
        "--ignore-status",
        type=int,
        nargs="*",
        help="ignore these HTTP status codes",
        metavar="HTTP status code",
        default=[],
    )
    parser.add_argument(
        "--warn-http", action="store_true", help="show warning for http links"
    )
    parser.add_argument(
        "--with-badge",
        action="store_true",
        help="generate a badge info file in json format"
        "(can be passed to linkmedkit scripts or shields.io API)",
    )
    parser.add_argument(
        "--port", type=int, help="webserver's port (default=8080)", default=8080
    )
    parser.add_argument(
        "--domain",
        help="name of the webserver's domain which its links will be treated as internal",
    )
    parser.add_argument(
        "--version", action="version", version="%(prog)s " + __version__
    )
    return parser.parse_args()



import logging
import warnings

import coloredlogs


# pylint: disable-next=unused-argument
def basic_warning_format(message, category, filename, lineno, *args, **kwargs):
    return "%s: %s" % (category.__name__, message)


# pylint: disable-next=unused-argument
def detailed_warning_format(message, category, filename, lineno, *args, **kwargs):
    return "%s:%s: %s: %s" % (filename, lineno, category.__name__, message)


def init_logger(logger_name, verbosity_level):
    """initialize the logger with a given name and verbosity

    :param logger_name:
    :param verbosity_level: 0,1,2,3(+)

    """
    logging.captureWarnings(True)
    warnings.formatwarning = basic_warning_format

    logger_handle = logging.getLogger(logger_name)
    logging_format = "%(levelname)-7s %(message)s"
    if verbosity_level == 0:
        logging.disable(logging.CRITICAL)
    else:
        if verbosity_level == 1:
            coloredlogs_level = "error"
            logging_level = logging.ERROR
        elif verbosity_level == 2:
            coloredlogs_level = "info"
            logging_format = "%(levelname)-7s %(name)-11s %(message)s"
            logging_level = logging.INFO
        else:
            coloredlogs_level = "debug"
            logging_format = "%(levelname)-7s %(name)-22s %(message)s"
            logging_level = logging.DEBUG
            warnings.formatwarning = detailed_warning_format

        logging.getLogger("urllib3").setLevel(logging_level)
        logger_handle.debug(
            'Initialized logger "%s" with level "%s"', logger_name, logging_level
        )
        coloredlogs.install(level=coloredlogs_level, fmt=logging_format)
        logger_handle.debug('Installing coloredlogs with level "%s"', coloredlogs_level)
    return logger_handle



import os
import re
import sys
import tempfile
import zipfile

from linkmedic.__init__ import __version__
from linkmedic.diagnose import diagnose
from linkmedic.logbook import init_logger
from linkmedic.reception import Orders, cli_args
from linkmedic.webserver import start_webserver


def main():
    ARGS = cli_args()
    do = Orders(ARGS)

    verbosity_level = 2 + ARGS.verbose
    if ARGS.quiet:
        verbosity_level = 1
    if ARGS.silent:
        verbosity_level = 0

    logger = init_logger("root", verbosity_level)

    logger.debug("linkmedic version %s", __version__)
    logger.debug("============================================================")
    logger.debug("webserver's root folder    : %s", ARGS.root)
    logger.debug("webserver port             : %d", ARGS.port)
    logger.debug("webserver's domain name    : %s", ARGS.domain)
    logger.debug("entry page                 : %s", ARGS.entry)
    logger.debug("entry page (expanded)      : %s", os.path.expanduser(ARGS.entry))
    logger.debug("redirect missing pages     : %r", not ARGS.no_redirect)
    logger.debug("check external links       : %s", ARGS.check_external)
    logger.debug("ignore local missing pages : %s", ARGS.ignore_local)
    logger.debug("ignore HTTP status codes   : %s", ARGS.ignore_status)
    logger.debug("warn http links            : %s", ARGS.warn_http)
    logger.debug("write badge info file      : %s", ARGS.with_badge)
    logger.debug("quiet output               : %r", ARGS.quiet)
    logger.debug("silence the output         : %r", ARGS.silent)
    logger.debug("verbosity level            : %s", verbosity_level)
    logger.debug("exit zero                  : %r", ARGS.exit_zero)
    logger.debug("Ignoring the links to      : %s", do.ignore_urls)
    logger.debug("============================================================")

    if not os.path.isdir(ARGS.root):
        logger.error("The webserver root path not found: %s", ARGS.root)
        sys.exit(2)

    if re.match(r".*\.[f]?od[tsp]{1}$", ARGS.entry):
        logger.debug("Entry page is an Open Document file")
        if not do.external_check:
            logger.debug(
                "Activating external links checking for the Open Document file"
            )
            do.external_check = True

        if re.match(r".*\.od[tsp]{1}$", ARGS.entry):
            tmp_dir = tempfile.TemporaryDirectory()
            logger.debug(
                "Created tmp directory for extracting Open Document file: %s",
                tmp_dir.name,
            )
            try:
                with zipfile.ZipFile(
                    os.path.expanduser(os.path.join(ARGS.root, ARGS.entry)), "r"
                ) as odf_archive:
                    odf_archive.extract("content.xml", path=tmp_dir.name)
            except Exception as err:
                logger.error(str(err))
                sys.exit(2)

            ARGS.root = tmp_dir.name
            ARGS.entry = "content.xml"

    server = start_webserver(
        ARGS.port,
        ARGS.root,
        ARGS.domain,
        not ARGS.no_redirect,
        verbosity_level,
    )
    logger.debug("Test webserver URL: %s", server.root_url)

    EXIT_CODE = diagnose(server, ARGS.entry, do, verbosity_level)

    if not ARGS.exit_zero:
        sys.exit(EXIT_CODE)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

# Copyright 2021-2023 M. Farzalipour Tabriz, Max Planck Computing and Data Facility (MPCDF)
# All rights reserved.
#
# This software may be modified and distributed under the terms
# of the 3-Clause BSD License. See the LICENSE file for details.

""" 
By directly running this script, it tries to spawn a webserver on port 8080 serving files from "./" 
and redirecting non-existing {url} to {url}.html

  ./webserver.py

You can pass "SERVER_PORT" "SERVER_PATH" "REDIRECT" "VERBOSITY_LEVEL" variables as cli parameters.

You can also use this as a library and start it by calling ``start_webserver()`` python function
which takes care of reconfiguring and restarting the webserver in case it fails to start.
"""

import http.server
import os
import socket
import socketserver
import subprocess
import sys
import time

import requests

from linkmedic.logbook import init_logger


class MedicServer:
    """contains popen object for the server subprocess
    and information about the server configuration

    """

    online = False
    port = 0
    root_url = ""
    domain = ""
    redirect = False
    handle = None

    def __init__(self, port, root_url, redirect, domain):
        self.port = port
        self.root_url = root_url
        self.redirect = redirect
        self.domain = domain

    def __del__(self):
        if self.handle:
            self.handle.terminate()

    def relative_address(self, url: str):
        """returns address relative to the website root

        :param url:

        """
        return url.replace(self.root_url, "")


def start_webserver(
    requested_port: int = 8080,
    server_root_path: str = "./",
    domain: str = "",
    redirect: bool = True,
    verbosity_level: int = 3,
):
    """Start a webserver and return the address to its root

    :param requested_port: requested port (may or may not be available)
    :returns: an instance of MedicServer

    """
    newserver = MedicServer(requested_port, server_root_path, redirect, domain)
    server_connection_timeout = 0.5
    CONNECTION_TIMEOUT_MAX = 5
    logger = init_logger("webmaster", verbosity_level)

    exec_dir = os.path.dirname(os.path.abspath(__file__))

    # Try to start a webserver
    while not newserver.online:
        logger.debug("Checking port %d", newserver.port)
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as test_socket:
            test_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            if test_socket.connect_ex(("localhost", newserver.port)) == 0:
                logger.info("Port %d is not avaiable!", newserver.port)
                newserver.port += 1
                continue
            logger.debug("Port %d seems to be free!", newserver.port)

        newserver.root_url = "http://localhost:" + str(newserver.port)
        logger.info("Starting test webserver on port %d", newserver.port)
        newserver.handle = subprocess.Popen(
            [
                "python3",
                os.path.join(exec_dir, "webserver.py"),
                str(newserver.port),
                server_root_path,
                str(newserver.redirect),
                str(verbosity_level),
            ],
        )
        time.sleep(server_connection_timeout)

        try:
            server_exit = newserver.handle.poll()
            if server_exit:
                raise ChildProcessError(
                    "Webserver exited unexpectedly! Exit code: " + str(server_exit)
                )
            requests.get(newserver.root_url, timeout=server_connection_timeout)
            logger.info("Webserver started!")
            newserver.online = True
        except Exception as startup_err:
            logger.debug(startup_err)
            newserver.handle.terminate()
            newserver.port += 1
            if server_connection_timeout < CONNECTION_TIMEOUT_MAX:
                server_connection_timeout += 0.1
                logger.debug(
                    "Connection timeout adjusted to %0.1f s", server_connection_timeout
                )

    return newserver


class MyHttpRequestHandler(http.server.SimpleHTTPRequestHandler):
    """ """

    # pylint: disable-next=redefined-builtin
    def log_message(self, format, *args):
        """log webserver messages only at debug

        :param format:
        :param *args:

        """
        logger.debug(format, *args)

    def do_GET(self):
        """selectively redirect missing pages to .html"""
        if REDIRECT:
            if not os.path.exists(os.getcwd() + self.path) and os.path.exists(
                os.getcwd() + self.path + ".html"
            ):
                logger.debug("Redirecting: %s -> %s.html", self.path, self.path)
                self.path += ".html"
        return http.server.SimpleHTTPRequestHandler.do_GET(self)


if __name__ == "__main__":
    SERVER_PORT = int(sys.argv[1]) if sys.argv[1:] else 8080
    SERVER_PATH = sys.argv[2] if sys.argv[2:] else "./"
    REDIRECT = sys.argv[3] == "True" if sys.argv[3:] else False
    VERBOSITY_LEVEL = int(sys.argv[4]) if sys.argv[4:] else 3

    logger = init_logger("webserver", VERBOSITY_LEVEL)

    os.chdir(SERVER_PATH)
    logger.debug("Requested webserver port       : %s", SERVER_PORT)
    logger.debug("Requested webserver path       : %s", SERVER_PATH)
    logger.debug("Redirect missing pages to .html: %r", REDIRECT)

    try:
        socketserver.TCPServer.allow_reuse_address = True
        with socketserver.TCPServer(("", SERVER_PORT), MyHttpRequestHandler) as server:
            server.serve_forever()
    except Exception as server_err:
        logger.debug(server_err)
        sys.exit(1)



# Copyright 2021-2023 M. Farzalipour Tabriz, Max Planck Computing and Data Facility (MPCDF)
# All rights reserved.
#
# This software may be modified and distributed under the terms
# of the 3-Clause BSD License. See the LICENSE file for details.


import json
import logging
import os
import re
from http import HTTPStatus
from urllib.parse import urljoin

import requests
from bs4 import BeautifulSoup, SoupStrainer

from linkmedic.logbook import init_logger
from linkmedic.reception import Orders
from linkmedic.webserver import MedicServer


class LinksStatus:
    """Aggregated statistics on status of tested links."""

    tested_links = 0
    dead_links = 0
    ignored_dead_links = 0
    http_links = 0
    parsed_urls = set()

    def __add__(self, other):
        total = self
        total.tested_links += other.tested_links
        total.dead_links += other.dead_links
        total.ignored_dead_links += other.ignored_dead_links
        total.http_links += other.http_links
        total.parsed_urls.update(other.parsed_urls)
        return total

    def __str__(self):
        parsed_links = len(self.parsed_urls)
        alive_links = self.tested_links - self.dead_links - self.ignored_dead_links

        stat_line = (
            "Discovered unique links: "
            + str(parsed_links)
            + " | "
            + str(self.tested_links)
            + " Tested | "
            + str(alive_links)
            + " Alive | "
            + str(self.dead_links)
            + " Dead ("
            + str(self.ignored_dead_links)
            + " ignored) || "
            + str(self.http_links)
            + " HTTP WARNING"
        )

        return stat_line

    def write_badge_files(self, tests_done: Orders):
        """write json badge description files

        :param tests_done: test orders performed

        """
        if not tests_done.internal_check:
            badge_name = "dead_external_links"
        elif not tests_done.external_check:
            badge_name = "dead_internal_links"
        else:
            badge_name = "dead_links"

        write_json_badge(badge_name, self.dead_links)

        if tests_done.http_check:
            http_badge_name = "http_links"
            write_json_badge(http_badge_name, self.http_links, critical=False)


def fetch_url(url: str):
    """

    :param url:
    :returns: http request response

    """
    headers = {
        "User-Agent": "Mozilla/5.0 (Windows NT 10.0; rv:102.0) Gecko/20100101 Firefox/102.0"
    }
    return requests.get(url, headers=headers, timeout=5)


def check_life(server: MedicServer, url: str, order: Orders):
    """check status of a url

    :param url:
    :returns: alive, msg, ignored_status

    """
    logger = logging.getLogger("diagnose")
    ignored_status = False
    dead_links_http_status_default = set([400, 404, 403, 408, 409, 501, 502, 503])
    logger.debug("Checking: %s", url)

    if order.ignore_status is not None:
        dead_links_http_status = dead_links_http_status_default.difference(
            order.ignore_status
        )
    else:
        dead_links_http_status = dead_links_http_status_default

    try:
        response = fetch_url(url)
        if response.status_code in dead_links_http_status:
            msg = (
                "DEAD LINK ("
                + str(response.status_code)
                + "-"
                + HTTPStatus(response.status_code).phrase
                + ") -> "
                + server.relative_address(url)
            )
            alive = False
        else:
            if response.status_code in order.ignore_status:
                msg = (
                    "Got response "
                    + str(response.status_code)
                    + " from the server which will be ignored! -> "
                    + server.relative_address(url)
                )
                ignored_status = True
            else:
                msg = "OK! -> " + server.relative_address(url)
            alive = True
        return alive, msg, ignored_status
    except requests.exceptions.RequestException as requests_err:
        logger.debug(requests_err)
        msg = "CONNECTION FAILED -> " + server.relative_address(url)
        return False, msg, False


def link_cleanup(link: str, server: MedicServer, base_address: str) -> str:
    """

    :param link:
    :param base_address: the page in which this link has been encountered e.g. https://abc.de/here

    """
    if re.match("^#", link):  # anchor in the base_address
        return ""
    if re.match("^mailto:", link):
        return ""
    if re.match(r".*(#)\S+", link):  # page + anchor
        link = re.search(r"(.*)#\S+", link).group(1)
    if not re.match(r"^https?:\/{2}", link):  # relative to the base_address
        link = urljoin(base_address, link)

    if server.domain:
        if re.match(r"^https?:\/{2}" + server.domain + "(.*)", link):
            internal_path = re.search(
                r"^https?:\/{2}" + server.domain + "(.*)", link
            ).group(1)
            link = urljoin(server.root_url, internal_path)

    return link


def crawl_url(url: str, server: MedicServer, order: Orders):
    """parses the url and checks all it links.

    returns LinksStatus, discovered_pages

    :param url:

    """
    logger = logging.getLogger("diagnose")
    stats = LinksStatus()
    discovered_pages = set()
    html_link_tags = {
        "a": "href",
        "iframe": "src",
        "img": "src",
        "link": "href",
        "script": "src",
    }
    xml_link_tags = {
        "a": "xlink:href",
        "event-listener": "xlink:href",
    }
    no_follow_exts = (
        ".conf",
        ".gif",
        ".ico",
        ".jpeg",
        ".jpg",
        ".js",
        ".pdf",
        ".png",
        ".svg",
        ".tar.gz",
        ".txt",
        ".woff2",
        ".zip",
    )

    try:
        response = fetch_url(url)
        base_address = response.url  # handle redirects
        if base_address.endswith(".xml") or re.match(r".*\.fod[tsp]{1}$", base_address):
            content_parser = "xml"
            link_tags = xml_link_tags
        else:
            content_parser = "html.parser"
            link_tags = html_link_tags

        for link_tag, link_url in iter(link_tags.items()):
            for link_obj in BeautifulSoup(
                response.text, content_parser, parse_only=SoupStrainer(link_tag)
            ):
                if link_obj.has_attr(link_url):
                    new_url = link_cleanup(link_obj[link_url], server, base_address)
                    if new_url:
                        logger.debug(
                            "Parser returned a link: %s",
                            server.relative_address(new_url),
                        )
                        if new_url not in stats.parsed_urls:
                            stats.parsed_urls.add(new_url)
                            logger.debug(
                                "New link to check: %s",
                                server.relative_address(new_url),
                            )
                            alive = None
                            ignored = False
                            ignored_status = False
                            msg = ""
                            if (
                                server.relative_address(new_url)
                                not in order.ignore_urls
                            ):
                                if server.root_url in new_url:  # local links
                                    alive, msg, ignored_status = check_life(
                                        server, new_url, order
                                    )
                                    stats.tested_links += 1
                                    if (
                                        not alive
                                        and order.internal_check
                                        and not ignored_status
                                    ):
                                        stats.dead_links += 1
                                    if not alive and not order.internal_check:
                                        stats.ignored_dead_links += 1
                                        ignored = True
                                        msg = (
                                            "IGNORED DEAD INTERNAL LINK -> "
                                            + server.relative_address(new_url)
                                        )

                                    if not new_url.endswith(no_follow_exts):
                                        logger.debug(
                                            "Links in this page should be followed -> %s",
                                            server.relative_address(new_url),
                                        )
                                        discovered_pages.add(new_url)

                                    else:
                                        logger.debug(
                                            "Links in this page will be ignored -> %s",
                                            server.relative_address(new_url),
                                        )
                                else:  # external links
                                    if order.external_check:
                                        alive, msg, ignored_status = check_life(
                                            server, new_url, order
                                        )
                                        stats.tested_links += 1
                                        if not alive and not ignored_status:
                                            stats.dead_links += 1
                                        if order.http_check:
                                            if "http://" in new_url:
                                                http_msg = "HTTP LINK -> " + new_url
                                                log_status(
                                                    http_msg,
                                                    "yellow",
                                                    server.relative_address(
                                                        base_address
                                                    ),
                                                )
                                                stats.http_links += 1
                                    else:
                                        ignored = True
                                        alive = True
                                        msg = "IGNORED EXTERNAL LINK -> " + new_url

                            else:  # url in ignore list
                                ignored = True
                                alive = True
                                msg = (
                                    "PAGE IN IGNORE LIST -> "
                                    + server.relative_address(new_url)
                                )
                            if not alive and ignored_status:
                                stats.ignored_dead_links += 1
                                msg_color = "yellow"
                            if ignored:
                                msg_color = "green" if alive else "yellow"
                            else:
                                msg_color = "green" if alive else "red"
                            log_status(
                                msg, msg_color, server.relative_address(base_address)
                            )
                        else:
                            logger.debug(
                                "Link was discovered before: %s",
                                server.relative_address(new_url),
                            )
        return stats, discovered_pages
    except Exception as crawler_err:
        logger.error("%s : %s", url, crawler_err)
        stats.dead_links += 1
        return stats, discovered_pages


def log_status(msg: str, color: str, page_url: str):
    """

    :param msg: param color:
    :param page_url: url for the page which cointains the link
    :param color:

    """
    logger = logging.getLogger("diagnose")
    if not msg:
        logger.warning("Internal error! log_status() was called with an empty message.")
        return

    if color == "green":
        msg_level = 10
    elif color == "yellow":
        msg_level = 30
    else:
        msg_level = 40

    msg = "@" + page_url + " > " + msg

    logger.log(msg_level, msg)


def write_json_badge(name: str, val: int, critical: bool = True):
    """

    :param name:
    :param val:
    :param critical: Default value = True

    """
    logger = logging.getLogger("diagnose")
    file = "badge." + name + ".json"
    badge_label = name.replace("_", "%20")

    if os.path.isfile(file):
        logger.debug("Overwriting the old badge info file: %s", file)
    else:
        logger.debug("Writing badge info file: %s", file)

    if val == 0:
        badge_color = "green"
    else:
        if critical:
            badge_color = "red"
        else:
            badge_color = "yellow"

    badge_data = {
        "schemaVersion": 1,
        "label": badge_label,
        "message": val,
        "color": badge_color,
    }
    with open(file, "w", encoding="utf-8") as json_file:
        json.dump(badge_data, json_file, ensure_ascii=False, indent=4)


def diagnose(server: MedicServer, entry: str, order: Orders, verbosity_level) -> int:
    ENTRY_URL = server.root_url + "/" + entry
    uncrawled_urls = set()
    stats = LinksStatus()
    logger = init_logger("diagnose", verbosity_level)
    logger.debug("Adding entry page's URL to the list: %s", ENTRY_URL)
    uncrawled_urls.add(ENTRY_URL)

    logger.info("Starting the crawler")
    try:
        # the entry page is checked separately!
        stats.parsed_urls.add(ENTRY_URL)
        alive, msg, _ = check_life(server, ENTRY_URL, order)
        stats.tested_links += 1
        msg_color = "green" if alive else "red"
        log_status(msg, msg_color, "$")
        if not alive:
            stats.dead_links = 1
            raise FileNotFoundError(
                "The entry page could not be loaded: "
                + server.relative_address(ENTRY_URL)
            )

        while uncrawled_urls:
            new_url = uncrawled_urls.pop()
            new_stats, extracted_urls_list = crawl_url(new_url, server, order)
            stats += new_stats
            if extracted_urls_list:
                logger.debug("Crawler extracted urls: %s", str(extracted_urls_list))
                uncrawled_urls.update(extracted_urls_list)

        logger.info("Crawler finished checking pages")

        if order.write_badge:
            logger.info("Writing badge files")
            stats.write_badge_files(order)
        logger.info(stats)
        RETURN_CODE = int(stats.dead_links > 0)
    except Exception as err:
        logger.error(err)
        RETURN_CODE = 2

    return RETURN_CODE
